/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
 
/*!
 *  @header SLPDARegisterer
 */

#ifndef _SLPDARegisterer_
#define _SLPDARegisterer_
#pragma once

#include <DirectoryServiceCore/DSLThread.h>

enum ActionType {
    kRegLocalService,
    kDeregLocalService,
    kRegisterAllServices,
    kRegisterAllServicesWithNewDA
};
 
struct DAInfoObj
{
    struct sockaddr_in 	sinDA;
    char* 				scopeList;
};

class RegData
{
public:
                RegData( const char* urlPtr, UInt32 urlLen, const char* scopeListPtr, UInt32 scopeListLen, const char* attributeListPtr, UInt32 attributeListLen );
                ~RegData();
                
    char*		mURLPtr;
    char*		mScopeListPtr;
    char*		mScopeList;
    char*		mAttributeListPtr;
    UInt32		mURLLen;
    UInt32		mScopeListLen;
    UInt32		mAttributeListLen;
};

void RegisterNewService( RegData* newReg );
void DeregisterService( RegData* newReg );

struct RegistrationObject
{
    ActionType	action;
    void*		data;
};

class SLPDARegisterer : public DSLThread
{
public:
    SLPDARegisterer( SLPHandle serverState );
    ~SLPDARegisterer();
    
    virtual void*		Run();
    
    void				RegisterAllServices( void ) { mRegisterAllServices = true; }
    void				AddTask( RegistrationObject* newTask );
    
    static void			QueueLock( void ) { pthread_mutex_lock( &mQueueLock ); }
    static void			QueueUnlock( void ) { pthread_mutex_unlock( &mQueueLock ); }
    static pthread_mutex_t		mQueueLock;

protected:
	SAState*			mServerState;
	SLPHandle			mSLPSA;
    CFMutableArrayRef	mActionQueue;
    Boolean				mClearQueue;
    Boolean				mRegisterAllServices;
    Boolean				mRegFileNeedsProcessing;
    Boolean				mCanceled;
};

#endif
