/*
 * Copyright (c) 2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
 
/*!
 *  @header SLPdSystemConfiguration
 */

#include <CoreFoundation/CoreFoundation.h>
#include <SystemConfiguration/SystemConfiguration.h>
#include <SystemConfiguration/SCDynamicStoreCopyDHCPInfo.h>
#include <SystemConfiguration/SCDynamicStorePrivate.h>

#ifndef _SLPdSystemConfiguration_
#define _SLPdSystemConfiguration_

class SLPdSystemConfiguration : public SLPSystemConfiguration
{
public:
                                    SLPdSystemConfiguration								( CFRunLoopRef runLoopRef = 0 );
                                    ~SLPdSystemConfiguration							();
                        
    static 	SLPSystemConfiguration*	TheSLPSC											( CFRunLoopRef runLoopRef = 0 );			// accessor to the class
    static	void					FreeSLPSC											( void );			

    virtual void					HandleIPv4Notification								( void );
	virtual void					HandleInterfaceNotification							( void );
private:
};

#endif
